package ij.gui;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Event;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Polygon;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.Shape;
import java.awt.event.KeyEvent;
import java.awt.geom.Path2D;
import java.awt.geom.GeneralPath;
import java.awt.geom.Rectangle2D;
import java.awt.geom.RoundRectangle2D;

import ij.IJ;
import ij.ImagePlus;
import ij.Undo;
import ij.WindowManager;
import ij.plugin.RectToolOptions;
import ij.plugin.frame.Recorder;
import ij.macro.Interpreter;
import ij.measure.Calibration;
import ij.plugin.OverlayCommands;
import ij.process.Blitter;
import ij.process.FloatPolygon;
import ij.process.ImageProcessor;


/** A rectangular region of interest and superclass for the other ROI classes. */
public class Roi extends AttributeRoi implements Cloneable, java.io.Serializable {
    //public class Roi extends AbstractRoi implements Cloneable, RoiShape, java.io.Serializable {


        int startX, startY, x, y, width, height;
        Rectangle2D.Double bounds, rect;
        public Shape roundRect;
        RoundRectangle2D.Double roundrect;
        double startXd, startYd, xd, yd, widthd, heightd;
        int activeHandle;
        int state;
        int modState = NO_MODS;
        int cornerDiameter;
        double cornerDiameterd;

        public static Roi previousRoi;
        public static final BasicStroke onePixelWide = new BasicStroke(1);
        //protected static Color ROIColor = Prefs.getColor(Prefs.ROICOLOR,Color.yellow);
        protected static int pasteMode = Blitter.COPY;
        protected static int lineWidth = 1;
        protected static Color defaultFillColor;

        protected int imageID;
        protected ImageCanvas ic;
        protected int oldX, oldY, oldWidth, oldHeight;
        protected int clipX, clipY, clipWidth, clipHeight;
        protected double oldXd, oldYd, oldWidthd, oldHeightd;
        protected double clipXd, clipYd, clipWidthd, clipHeightd;
        protected ImagePlus clipboard;
        protected boolean constrain; // to be square
        protected boolean center;
        protected boolean aspect;
        protected boolean updateFullWindow;
        protected double mag = 1.0;
        protected double asp_bk; //saves aspect ratio if resizing takes roi very small
        protected ImageProcessor cachedMask;
        protected Color handleColor = Color.white;
        protected Color instanceColor; //obsolete; replaced by  strokeColor
        protected boolean nonScalable;
        protected boolean overlay;
        private Overlay prototypeOverlay;
        private boolean subPixel;
        private static final double acv;


        static {
            double angle = Math.PI / 4.0;
            double a = 1.0 - Math.cos(angle);
            double b = Math.tan(angle);
            double c = Math.sqrt(1.0 + b * b) - 1 + a;
            double cv = 4.0 / 3.0 * a * b / c;
            acv = (1.0 - cv);
        }

        /** Creates a rectangular ROI. */
        public Roi(int x, int y, int width, int height) {
                this(x, y, width, height, 0);
        }


        /** Creates a new rounded rectangular ROI. */
        public Roi(int x, int y, int width, int height, int cornerDiameter) {
                setImage(null);
                if (width<1) width = 1;
                if (height<1) height = 1;
                if (width>xMax) width = xMax;
                if (height>yMax) height = yMax;
                this.cornerDiameter = cornerDiameter;
                //setLocation(x, y);
                this.x = x;
                this.y = y;
                startX = x; startY = y;
                oldX = x; oldY = y; oldWidth=0; oldHeight=0;
                this.width = width;
                this.height = height;
                oldWidth=width;
                oldHeight=height;
                clipX = x;
                clipY = y;
                clipWidth = width;
                clipHeight = height;
                state = NORMAL;
                type = RECTANGLE;
                if (ic!=null) {
                        Graphics g = ic.getGraphics();
                        draw(g);
                        g.dispose();
                }
                fillColor = defaultFillColor;
                xd=x; yd=y; widthd=width; heightd=height;
        }
        /** Creates a new rounded rectangular ROI. */
        public Roi(double x, double y, double width, double height, double cornerDiameter) {
                setImage(null);
                if (width<1) width = 1;
                if (height<1) height = 1;
                if (width>xMax) width = xMax;
                if (height>yMax) height = yMax;
                this.cornerDiameterd = cornerDiameter;
                //setLocation(x, y);
                this.xd = x;
                this.yd = y;
                startXd = x; startYd = y;
                oldXd = x; oldYd = y; oldWidthd=0; oldHeightd=0;
                this.widthd = width;
                this.heightd = height;
                oldWidthd=width;
                oldHeightd=height;
                clipXd = x;
                clipYd = y;
                clipWidthd = width;
                clipHeightd = height;
                state = NORMAL;
                type = RECTANGLE;
                if (ic!=null) {
                        Graphics g = ic.getGraphics();
                        draw(g);
                        g.dispose();
                }
                fillColor = defaultFillColor;
                xd=x; yd=y; widthd=width; heightd=height;
                //subPixel = true;
        }

        /** Creates a rounded rectangular ROI using double arguments. */
        public Roi(double x, double y, double width, double height) {
                this(x, y, Math.ceil(width), Math.ceil(height), 0);
                xd=x; yd=y; widthd=width; heightd=height;
                subPixel = true;
        }

        /** Creates a new rectangular Roi. */
        public Roi(Rectangle r) {
                this(r.x, r.y, r.width, r.height);
        }
        /** Creates a new rectangular Roi. */
        public Roi(Rectangle2D.Double r) {
                this(r.x, r.y, r.width, r.height);
        }

        /** Starts the process of creating a user-defined rectangular Roi,
                where sx and sy are the starting screen coordinates. */
        public Roi(int sx, int sy, ImagePlus imp) {
                this(sx, sy, imp, 0);
        }

        /** Starts the process of creating a user-defined rectangular Roi,
                where sx and sy are the starting screen coordinates. */
        public Roi(int sx, int sy, ImagePlus imp, int cornerDiameter) {
                setImage(imp);
                int ox=sx, oy=sy;
                if (ic!=null) {
                        ox = ic.offScreenX(sx);
                        oy = ic.offScreenY(sy);
                }
                setLocation(ox, oy);
                this.cornerDiameter = cornerDiameter;
                width = 0;
                height = 0;
                state = CONSTRUCTING;
                type = RECTANGLE;
                if (cornerDiameter>0) {
                        double swidth = RectToolOptions.getDefaultStrokeWidth();
                        if (swidth>0.0)
                                setStrokeWidth(swidth);
                        Color scolor = RectToolOptions.getDefaultStrokeColor();
                        if (scolor!=null)
                                setStrokeColor(scolor);
                }
                fillColor = defaultFillColor;
        }

        /** @deprecated */
        public Roi(int x, int y, int width, int height, ImagePlus imp) {
                this(x, y, width, height);
                setImage(imp);
        }

        /** Set the location of the ROI in image coordinates. */
        public void setLocation(int x, int y) {
                this.x = x;
                this.y = y;
                startX = x; startY = y;
                oldX = x; oldY = y; oldWidth=0; oldHeight=0;
                xd=x; yd=y;
        }

        public void setImage(ImagePlus imp) {
                this.imp = imp;
                cachedMask = null;
                if (imp==null) {
                        ic = null;
                        clipboard = null;
                        xMax = 99999;
                        yMax = 99999;
                } else {
                        ic = imp.getCanvas();
                        xMax = imp.getWidth();
                        yMax = imp.getHeight();
                }
        }

        public int getState() {
                return state;
        }

        /** Returns the perimeter length. */
        public double getLength() {
                double pw=1.0, ph=1.0;
                if (imp!=null) {
                        Calibration cal = imp.getCalibration();
                        pw = cal.pixelWidth;
                        ph = cal.pixelHeight;
                }
                return 2.0*width*pw+2.0*height*ph;
        }



        /** Return this selection's bounding rectangle. */
        public Rectangle getBounds() {
                return new Rectangle(x, y, width, height);
        }

        /** Return this selection's bounding rectangle. */
        public Rectangle2D.Double getFloatBounds() {
                if (bounds!=null)
                        return new Rectangle2D.Double(bounds.x, bounds.y, bounds.width, bounds.height);
                else
                        return new Rectangle2D.Double(x, y, width, height);
        }

        /**
        * @deprecated
        * replaced by getBounds()
        */
        public Rectangle getBoundingRect() {
                return getBounds();
        }

        /** Returns the outline of this selection as a Polygon, or
                null if this is a straight line selection.
                @see ij.process.ImageProcessor#setRoi
                @see ij.process.ImageProcessor#drawPolygon
                @see ij.process.ImageProcessor#fillPolygon
        */
        public Polygon getPolygon() {
                int[] xpoints = new int[4];
                int[] ypoints = new int[4];
                xpoints[0] = x;
                ypoints[0] = y;
                xpoints[1] = x+width;
                ypoints[1] = y;
                xpoints[2] = x+width;
                ypoints[2] = y+height;
                xpoints[3] = x;
                ypoints[3] = y+height;
                return new Polygon(xpoints, ypoints, 4);
        }

        public FloatPolygon getFloatPolygon() {
                if (subPixelResolution()) {
                        float[] xpoints = new float[4];
                        float[] ypoints = new float[4];
                        xpoints[0] = (float)xd;
                        ypoints[0] = (float)yd;
                        xpoints[1] = (float)(xd+widthd);
                        ypoints[1] = (float)yd;
                        xpoints[2] = (float)(xd+widthd);
                        ypoints[2] = (float)(yd+heightd);
                        xpoints[3] = (float)xd;
                        ypoints[3] = (float)(yd+heightd);
                        return new FloatPolygon(xpoints, ypoints);
                } else {
                        Polygon p = getPolygon();
                        return new FloatPolygon(toFloat(p.xpoints), toFloat(p.ypoints), p.npoints);
                }
        }

        protected void grow(int sx, int sy) {
                if (clipboard!=null) return;
                int xNew = ic.offScreenX(sx);
                int yNew = ic.offScreenY(sy);
                if (type==RECTANGLE) {
                        if (xNew < 0) xNew = 0;
                        if (yNew < 0) yNew = 0;
                }
                if (constrain) {
                        // constrain selection to be square
                        if (!center)
                                {growConstrained(xNew, yNew); return;}
                        int dx, dy, d;
                        dx = xNew - x;
                        dy = yNew - y;
                        if (dx<dy)
                                d = dx;
                        else
                                d = dy;
                        xNew = x + d;
                        yNew = y + d;
                }
                if (center) {
                        width = Math.abs(xNew - startX)*2;
                        height = Math.abs(yNew - startY)*2;
                        x = startX - width/2;
                        y = startY - height/2;
                } else {
                        width = Math.abs(xNew - startX);
                        height = Math.abs(yNew - startY);
                        x = (xNew>=startX)?startX:startX - width;
                        y = (yNew>=startY)?startY:startY - height;
                        if (type==RECTANGLE) {
                                if ((x+width) > xMax) width = xMax-x;
                                if ((y+height) > yMax) height = yMax-y;
                        }
                }
                updateClipRect();
                imp.draw(clipX, clipY, clipWidth, clipHeight);
                oldX = x;
                oldY = y;
                oldWidth = width;
                oldHeight = height;
                xd=x; yd=y; widthd=width; heightd=height;
        }

        private void growConstrained(int xNew, int yNew) {
                int dx = xNew - startX;
                int dy = yNew - startY;
                width = height = (int)Math.round(Math.sqrt(dx*dx + dy*dy));
                if (type==RECTANGLE) {
                        x = (xNew>=startX)?startX:startX - width;
                        y = (yNew>=startY)?startY:startY - height;
                        if (x<0) x = 0;
                        if (y<0) y = 0;
                        if ((x+width) > xMax) width = xMax-x;
                        if ((y+height) > yMax) height = yMax-y;
                } else {
                        x = startX + dx/2 - width/2;
                        y = startY + dy/2 - height/2;
                }
                updateClipRect();
                imp.draw(clipX, clipY, clipWidth, clipHeight);
                oldX = x;
                oldY = y;
                oldWidth = width;
                oldHeight = height;
        }

        protected void moveHandle(int sx, int sy) {
                double asp;
                if (clipboard!=null) return;
                int ox = ic.offScreenX(sx);
                int oy = ic.offScreenY(sy);
                if (ox<0) ox=0; if (oy<0) oy=0;
                if (ox>xMax) ox=xMax; if (oy>yMax) oy=yMax;
                //IJ.log("moveHandle: "+activeHandle+" "+ox+" "+oy);
                int x1=x, y1=y, x2=x1+width, y2=y+height, xc=x+width/2, yc=y+height/2;
                if (width > 7 && height > 7) {
                        asp = (double)width/(double)height;
                        asp_bk = asp;
                } else {
                        asp = asp_bk;
                }

                //if(activeHandle==0)return;
                switch (activeHandle) {
                        case 0:
                                x=ox; y=oy;
                                break;
                        case 1:
                                y=oy;
                                break;
                        case 2:
                                x2=ox; y=oy;
                                break;
                        case 3:
                                x2=ox;
                                break;
                        case 4:
                                x2=ox; y2=oy;
                                break;
                        case 5:
                                y2=oy;
                                break;
                        case 6:
                                x=ox; y2=oy;
                                break;
                        case 7:
                                x=ox;
                                break;
                }
                if (x<x2)
                   width=x2-x;
                else
                  {width=1; x=x2;}
                if (y<y2)
                   height = y2-y;
                else
                   {height=1; y=y2;}

                if (center) {
                        switch(activeHandle){
                                case 0:
                                        width=(xc-x)*2;
                                        height=(yc-y)*2;
                                        break;
                                case 1:
                                        height=(yc-y)*2;
                                        break;
                                case 2:
                                        width=(x2-xc)*2;
                                        x=x2-width;
                                        height=(yc-y)*2;
                                        break;
                                case 3:
                                        width=(x2-xc)*2;
                                        x=x2-width;
                                        break;
                                case 4:
                                        width=(x2-xc)*2;
                                        x=x2-width;
                                        height=(y2-yc)*2;
                                        y=y2-height;
                                        break;
                                case 5:
                                        height=(y2-yc)*2;
                                        y=y2-height;
                                        break;
                                case 6:
                                        width=(xc-x)*2;
                                        height=(y2-yc)*2;
                                        y=y2-height;
                                        break;
                                case 7:
                                        width=(xc-x)*2;
                                        break;
                        }
                        if (x>=x2) {
                                width=1;
                                x=x2=xc;
                        }
                        if (y>=y2) {
                                height=1;
                                y=y2=yc;
                        }
                        xd=x; yd=y;

                }

                if (constrain) {
                        if (activeHandle==1 || activeHandle==5)
                                width=height;
                        else
                                height=width;

                        if(x>=x2) {
                                width=1;
                                x=x2=xc;
                        }
                        if (y>=y2) {
                                height=1;
                                y=y2=yc;
                        }
                        switch(activeHandle){
                                case 0:
                                        x=x2-width;
                                        y=y2-height;
                                        break;
                                case 1:
                                        x=xc-width/2;
                                        y=y2-height;
                                        break;
                                case 2:
                                        y=y2-height;
                                        break;
                                case 3:
                                        y=yc-height/2;
                                        break;
                                case 5:
                                        x=xc-width/2;
                                        break;
                                case 6:
                                        x=x2-width;
                                        break;
                                case 7:
                                        y=yc-height/2;
                                        x=x2-width;
                                        break;
                        }
                        if (center){
                                x=xc-width/2;
                                y=yc-height/2;
                        }
                }

                if (aspect && !constrain) {
                        if (activeHandle==1 || activeHandle==5) width=(int)Math.rint((double)height*asp);
                        else height=(int)Math.rint((double)width/asp);

                        switch(activeHandle){
                                case 0:
                                        x=x2-width;
                                        y=y2-height;
                                        break;
                                case 1:
                                        x=xc-width/2;
                                        y=y2-height;
                                        break;
                                case 2:
                                        y=y2-height;
                                        break;
                                case 3:
                                        y=yc-height/2;
                                        break;
                                case 5:
                                        x=xc-width/2;
                                        break;
                                case 6:
                                        x=x2-width;
                                        break;
                                case 7:
                                        y=yc-height/2;
                                        x=x2-width;
                                        break;
                        }
                        if (center){
                                x=xc-width/2;
                                y=yc-height/2;
                        }

                        // Attempt to preserve aspect ratio when roi very small:
                        if (width<8) {
                                if(width<1) width = 1;
                                height=(int)Math.rint((double)width/asp_bk);
                        }
                        if (height<8) {
                                if(height<1) height =1;
                                width=(int)Math.rint((double)height*asp_bk);
                        }

                }

                updateClipRect();
                imp.draw(clipX, clipY, clipWidth, clipHeight);
                oldX=x; oldY=y;
                oldWidth=width; oldHeight=height;
        }

        void move(int sx, int sy) {
                int xNew = ic.offScreenX(sx);
                int yNew = ic.offScreenY(sy);
                x += xNew - startX;
                y += yNew - startY;
                boolean isImageRoi = this instanceof ImageRoi;
                if (clipboard==null && type==RECTANGLE && !isImageRoi) {
                        if (x<0) x=0; if (y<0) y=0;
                        if ((x+width)>xMax) x = xMax-width;
                        if ((y+height)>yMax) y = yMax-height;
                }
                startX = xNew;
                startY = yNew;
                updateClipRect();
                if (lineWidth>1 && isLine())
                        imp.draw();
                else
                        imp.draw(clipX, clipY, clipWidth, clipHeight);
                oldX = x;
                oldY = y;
                oldWidth = width;
                oldHeight=height;
                if (isImageRoi) showStatus();
                xd=x; yd=y;
                bounds = null;
        }

        /** Nudge ROI one pixel on arrow key press. */
        public void nudge(int key) {
                switch(key) {
                        case KeyEvent.VK_UP:
                                y--;
                                if (y<0 && (type!=RECTANGLE||clipboard==null))
                                        y = 0;
                                break;
                        case KeyEvent.VK_DOWN:
                                y++;
                                if ((y+height)>=yMax && (type!=RECTANGLE||clipboard==null))
                                        y = yMax-height;
                                break;
                        case KeyEvent.VK_LEFT:
                                x--;
                                if (x<0 && (type!=RECTANGLE||clipboard==null))
                                        x = 0;
                                break;
                        case KeyEvent.VK_RIGHT:
                                x++;
                                if ((x+width)>=xMax && (type!=RECTANGLE||clipboard==null))
                                        x = xMax-width;
                                break;
                }
                updateClipRect();
                imp.draw(clipX, clipY, clipWidth, clipHeight);
                oldX = x; oldY = y;
                xd=x; yd=y;
                showStatus();
        }

        /** Nudge lower right corner of rectangular and oval ROIs by
                one pixel based on arrow key press. */
        public void nudgeCorner(int key) {
                if (type>OVAL || clipboard!=null)
                        return;
                switch(key) {
                        case KeyEvent.VK_UP:
                                height--;
                                if (height<1) height = 1;
                                break;
                        case KeyEvent.VK_DOWN:
                                height++;
                                if ((y+height) > yMax) height = yMax-y;
                                break;
                        case KeyEvent.VK_LEFT:
                                width--;
                                if (width<1) width = 1;
                                break;
                        case KeyEvent.VK_RIGHT:
                                width++;
                                if ((x+width) > xMax) width = xMax-x;
                                break;
                }
                updateClipRect();
                imp.draw(clipX, clipY, clipWidth, clipHeight);
                oldX = x; oldY = y;
                cachedMask = null;
                showStatus();
        }

        protected void updateClipRect() {
        // Finds the union of current and previous roi
                clipX = (x<=oldX)?x:oldX;
                clipY = (y<=oldY)?y:oldY;
                clipWidth = ((x+width>=oldX+oldWidth)?x+width:oldX+oldWidth) - clipX + 1;
                clipHeight = ((y+height>=oldY+oldHeight)?y+height:oldY+oldHeight) - clipY + 1;
                int m = 3;
                if (ic!=null) {
                        double mag = ic.getMagnification();
                        if (mag<1.0)
                                m = (int)(4.0/mag);
                }
                m += clipRectMargin();
                m = (int)(m+getStrokeWidth()*2);
                clipX-=m; clipY-=m;
                clipWidth+=m*2; clipHeight+=m*2;
                //if (IJ.debugMode) IJ.log("updateClipRect: "+m+"  "+clipX+" "+clipY+" "+clipWidth+" "+clipHeight);
         }

        protected int clipRectMargin() {
                return 0;
        }

        protected void handleMouseDrag(int sx, int sy, int flags) {
                if (ic==null) return;
                constrain = (flags&Event.SHIFT_MASK)!=0;
                center = (flags&Event.CTRL_MASK)!=0 || (IJ.isMacintosh()&&(flags&Event.META_MASK)!=0);
                aspect = (flags&Event.ALT_MASK)!=0;
                switch(state) {
                        case CONSTRUCTING:
                                grow(sx, sy);
                                break;
                        case MOVING:
                                move(sx, sy);
                                break;
                        case MOVING_HANDLE:
                                moveHandle(sx, sy);
                                break;
                        default:
                                break;
                }
        }

        public int getHandleSize() {
                double mag = ic!=null?ic.getMagnification():1.0;
                double size = HANDLE_SIZE/mag;
                return (int)(size*mag);
        }

        public void draw(Graphics g) {
                Color color =  strokeColor!=null? strokeColor:ROIColor;
                if (fillColor!=null) color = fillColor;
                if (Interpreter.isBatchMode() && ic!=null && imp.getOverlay()!=null && strokeColor==null && fillColor==null)
                        return;
                g.setColor(color);
                mag = getMagnification();
                double sw = (width*mag);
                double sh = (height*mag);
                double sx1 = screenX(x);
                double sy1 = screenY(y);
                if (subPixelResolution()) {
                        sw = (widthd*mag);
                        sh = (heightd*mag);
                        sx1 = screenXD(xd);
                        sy1 = screenYD(yd);
                }
                double sx2 = sx1+sw/2;
                double sy2 = sy1+sh/2;
                double sx3 = sx1+sw;
                double sy3 = sy1+sh;
                Graphics2D g2d = (Graphics2D)g;
                double sArcSized = (double)Math.round(cornerDiameter*mag);
                rect = new Rectangle2D.Double(sx1, sy1, sw, sh);
                roundRect = new RoundRectangle2D.Double(sx1, sy1, sw, sh, sArcSized, sArcSized);
                if (stroke!=null)
                    g2d.setStroke(getScaledStroke());
                        g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
                    if (cornerDiameter>0) {
                        /*drawStroke(g2d);*/
                             //int sArcSize = (int)Math.round(cornerDiameter*mag);
                            if (fillColor!=null){
                                    //g.fillRoundRect(sx1, sy1, sw, sh, sArcSize, sArcSize);
                                    g2d.setColor(color);
                                    g2d.fill(roundRect);
                            }
                            else{
                                //g.drawRoundRect(sx1, sy1, sw, sh, sArcSize, sArcSize);
                                g2d.draw(roundRect);
                            }
                    } else {
                        if (fillColor!=null){
                                g2d.setColor(color);
                                g2d.fill(rect);
                        }
                        else{
                            g2d.draw(rect);
                        }
                }
                /**
                 * Draw Handle clockwise. start corner up-left  1,2,3,5,8,7,6,4
                 *
                 * */
                if (state!=CONSTRUCTING && clipboard==null && !overlay) {
                        int size2 = HANDLE_SIZE/2;
                        double sxe = sx1+sw/4;
                        double sye = sy1+sh/4;
                        drawHandle(g, (int)sx1-size2, (int)sy1-size2);
                        drawHandle(g, (int)sx2-size2, (int)sy1-size2);
                        drawHandle(g, (int)sx3-size2, (int)sy1-size2);
                        drawHandle(g, (int)sx3-size2, (int)sy2-size2);
                        drawHandle(g, (int)sx3-size2, (int)sy3-size2);
                        drawHandle(g, (int)sx2-size2, (int)sy3-size2);
                        drawHandle(g, (int)sx1-size2, (int)sy3-size2);
                        drawHandle(g, (int)sx1-size2, (int)sy2-size2);
                        fillRoiHandle(g, sxe, sy1);
                        drawRoiHandle(g, sx1, sye);
                }
                drawPreviousRoi(g);
                if (state!=NORMAL) showStatus();
                if (updateFullWindow)
                        {updateFullWindow = false; imp.draw();}
        }

        public void drawOverlay(Graphics g) {
                overlay = true;
                draw(g);
                overlay = false;
        }

        public void drawfillOverlay(Graphics g, boolean fillcolor) {
                overlay = true;
                if(this instanceof Line){
                    fillcolor=false;
                }
                else if(isHandle(screenX(x), screenY(y)) == DRAW_HANDLE){
                    fillcolor=false;
                }
                fillColor= fillcolor ? getColor(): null;//Toolbar.getForegroundColor();
                /*if((this instanceof Line)){
                    draw(g);
                }
                  else if(isHandle(screenX(x), screenY(y)) != FILL_HANDLE){
                     draw(g);
                     //OverlayCommands.addSelection();
                  }*/
                     OverlayCommands.drawfillShape();
                     fillColor= null;
                overlay = false;
        }

        void drawPreviousRoi(Graphics g) {
                if (previousRoi!=null && previousRoi!=this && previousRoi.modState!=NO_MODS) {
                        if (type!=POINT && previousRoi.getType()==POINT && previousRoi.modState!=SUBTRACT_FROM_ROI)
                                return;
                        previousRoi.setImage(imp);
                        previousRoi.draw(g);
                }
        }

        public void drawRoiHandle(Graphics g, double x, double y) {
            BasicStroke stroke = new BasicStroke(1f);
            Graphics2D g2d = (Graphics2D)g;
            Path2D.Double p2D = new Path2D.Double();
            p2D.moveTo(x-4, y);
            p2D.lineTo(x, y+4);
            p2D.lineTo(x+4, y);
            p2D.lineTo(x, y-4);
            p2D.lineTo(x-4, y);
            g2d.setColor(Color.black);
            g2d.setStroke(stroke);
            p2D.closePath();
            g2d.draw(p2D);

            p2D.moveTo(x-2, y);
            p2D.lineTo(x, y+2);
            p2D.lineTo(x+2, y);
            p2D.lineTo(x, y-2);
            p2D.lineTo(x-2, y);
            //g2d.setStroke(stroke);
            g2d.setColor(new Color(255, 102, 204));
            p2D.closePath();
            g2d.fill(p2D);
       }
        public void fillRoiHandle(Graphics g, double x, double y) {
            BasicStroke stroke = new BasicStroke(1f);
            Graphics2D g2d = (Graphics2D)g;
            Path2D.Double p2D = new Path2D.Double();
            p2D.moveTo(x-4, y);
            p2D.lineTo(x, y+4);
            p2D.lineTo(x+4, y);
            p2D.lineTo(x, y-4);
            p2D.lineTo(x-4, y);
            g2d.setColor(Color.black);
            g2d.setStroke(stroke);
            p2D.closePath();
            g2d.draw(p2D);

            p2D.moveTo(x-2, y);
            p2D.lineTo(x, y+2);
            p2D.lineTo(x+2, y);
            p2D.lineTo(x, y-2);
            p2D.lineTo(x-2, y);
            //g2d.setStroke(stroke);
            g2d.setColor(new Color(144, 222, 234));
            p2D.closePath();
            g2d.fill(p2D);
       }
            public void drawHandle(Graphics g, int x, int y) {
                double size = (width*height)*mag*mag;
                if (type==LINE) {
                        size = Math.sqrt(width*width+height*height);
                        size *= size*mag*mag;
                }
                if (size>4000.0) {
                        g.setColor(Color.black);
                        g.fillRect(x,y,5,5);
                        g.setColor(handleColor);
                        g.fillRect(x+1,y+1,3,3);
                } else if (size>1000.0) {
                        g.setColor(Color.black);
                        g.fillRect(x+1,y+1,4,4);
                        g.setColor(handleColor);
                        g.fillRect(x+2,y+2,2,2);
                } else {
                        g.setColor(Color.black);
                        g.fillRect(x+1,y+1,3,3);
                        g.setColor(handleColor);
                        g.fillRect(x+2,y+2,1,1);
                }
        }

        /**
        * @deprecated
        * replaced by drawPixels(ImageProcessor)
        */
        public void drawPixels() {
                if (imp!=null)
                        drawPixels(imp.getProcessor());
        }

        /** Draws the selection outline on the specified ImageProcessor.
                @see ij.process.ImageProcessor#setColor
                @see ij.process.ImageProcessor#setLineWidth
        */
        public void drawPixels(ImageProcessor ip) {
                endPaste();
                int saveWidth = ip.getLineWidth();
                Shape shape = new RoundRectangle2D.Float(x, y, width,
                            height, cornerDiameter, cornerDiameter);
                GeneralPath path = new GeneralPath(GeneralPath.WIND_EVEN_ODD);
                path.append(shape, true);
                if (getStrokeWidth()>1f)
                        ip.setLineWidth((int)Math.round(getStrokeWidth()));
                if (cornerDiameter>0){
                    new ShapeRoi(shape).drawPixels(ip, path);
                }
                else {
                        if (ip.getLineWidth()==1)
                                ip.drawRect(x, y, width+1, height+1);
                        else
                                ip.drawRect(x, y, width, height);
                }
                ip.setLineWidth(saveWidth);
                if (Line.getWidth()>1 || getStrokeWidth()>1)
                        updateFullWindow = true;
        }
        /** Draws the selection outline on the specified ImageProcessor.
                @see ij.process.ImageProcessor#setColor
                @see ij.process.ImageProcessor#setLineWidth
        */
        public void drawPixels(ImageProcessor ip, GeneralPath path) {
                endPaste();
                int saveWidth = ip.getLineWidth();
                Shape shape = new RoundRectangle2D.Float(x, y, width,
                            height, cornerDiameter, cornerDiameter);
                path = new GeneralPath(GeneralPath.WIND_EVEN_ODD);
                path.append(shape, true);
                if (getStrokeWidth()>1f)
                        ip.setLineWidth((int)Math.round(getStrokeWidth()));
                    if (cornerDiameter>0){
                        ShapeRoi sr =new ShapeRoi(shape);
                        sr.drawPixels(ip, path);
                    }
                else {
                        if (ip.getLineWidth()==1)
                                ip.drawRect(x, y, width+1, height+1);
                        else
                                ip.drawRect(x, y, width, height);
                }
                ip.setLineWidth(saveWidth);
                if (Line.getWidth()>1 || getStrokeWidth()>1)
                        updateFullWindow = true;
        }

        public void drawStroke(Graphics2D g) {
            //Graphics2D g = (Graphics2D)gs;
           if (roundrect.archeight == 0 && roundrect.arcwidth == 0) {
               g.draw(roundrect.getBounds2D());
           } else {
               // We have to generate the path for the round rectangle manually,
               // because the path of a Java RoundRectangle is drawn counter clockwise
               // whereas an SVG rect needs to be drawn clockwise.
               Path2D.Double p = new Path2D.Double();
               double aw = roundrect.arcwidth / 2d;
               double ah = roundrect.archeight / 2d;
               
               p.moveTo((roundrect.x + aw), (float) roundrect.y);
               
               p.lineTo((roundrect.x + roundrect.width - aw), (float) roundrect.y);
               p.curveTo((roundrect.x + roundrect.width - aw * acv), (float) roundrect.y, //
                       (roundrect.x + roundrect.width), (float)(roundrect.y + ah * acv), //
                       (roundrect.x + roundrect.width), (roundrect.y + ah));
               p.lineTo((roundrect.x + roundrect.width), (roundrect.y + roundrect.height - ah));
               p.curveTo(
                       (roundrect.x + roundrect.width), (roundrect.y + roundrect.height - ah * acv),//
                       (roundrect.x + roundrect.width - aw * acv), (roundrect.y + roundrect.height),//
                       (roundrect.x + roundrect.width - aw), (roundrect.y + roundrect.height));
               p.lineTo((roundrect.x + aw), (roundrect.y + roundrect.height));
               p.curveTo((roundrect.x + aw*acv), (roundrect.y + roundrect.height),//
                       (roundrect.x), (roundrect.y + roundrect.height - ah*acv),//
                      (float) roundrect.x, (roundrect.y + roundrect.height - ah));
               p.lineTo((float) roundrect.x, (roundrect.y + ah));
               p.curveTo((roundrect.x), (roundrect.y + ah*acv),//
                       (roundrect.x + aw*acv), (float)(roundrect.y),//
                       (float)(roundrect.x + aw), (float)(roundrect.y));
               p.closePath();
               g.draw(p);
           }
       }
       public boolean contains(int x, int y) {
                Rectangle r = new Rectangle(this.x, this.y, width, height);
                boolean contains = r.contains(x, y);
                if (cornerDiameter==0 || contains==false)
                        return contains;
                RoundRectangle2D rr = new RoundRectangle2D.Float(this.x, this.y, width, height, cornerDiameter, cornerDiameter);
                return rr.contains(x, y);
        }

        /** Returns a handle number if the specified screen coordinates are
                inside or near a handle, otherwise returns -1. */
        public int isHandle(int sx, int sy) {
                if (clipboard!=null || ic==null) return -1;
                double mag = ic.getMagnification();
                int size = HANDLE_SIZE+3;
                int halfSize = size/2;
                int sx1 = ic.screenX(x) - halfSize;
                int sy1 = ic.screenY(y) - halfSize;
                int sx3 = ic.screenX(x+width) - halfSize;
                int sy3 = ic.screenY(y+height) - halfSize;
                int sx2 = sx1 + (sx3 - sx1)/2;
                int sy2 = sy1 + (sy3 - sy1)/2;
                int sxe = ic.screenX(x+width/4);
                int sye = ic.screenY(y+height/4);
                if (sx>=sx1&&sx<=sx1+size&&sy>=sy1&&sy<=sy1+size) return 0;
                if (sx>=sx2&&sx<=sx2+size&&sy>=sy1&&sy<=sy1+size) return 1;
                if (sx>=sx3&&sx<=sx3+size&&sy>=sy1&&sy<=sy1+size) return 2;
                if (sx>=sx3&&sx<=sx3+size&&sy>=sy2&&sy<=sy2+size) return 3;
                if (sx>=sx3&&sx<=sx3+size&&sy>=sy3&&sy<=sy3+size) return 4;
                if (sx>=sx2&&sx<=sx2+size&&sy>=sy3&&sy<=sy3+size) return 5;
                if (sx>=sx1&&sx<=sx1+size&&sy>=sy3&&sy<=sy3+size) return 6;
                if (sx>=sx1&&sx<=sx1+size&&sy>=sy2&&sy<=sy2+size) return 7;
                if (sx>=sxe&&sx<=sxe+size&&sy>=sy1&&sy<=sy1+size) return FILL_HANDLE;
                if (sx>=sx1&&sx<=sx1+size&&sy>=sye&&sy<=sye+size) return DRAW_HANDLE;
                return -1;
        }

        protected void mouseDownInHandle(int handle, int sx, int sy) {
                state = MOVING_HANDLE;
                activeHandle = handle;
        }

        protected void handleMouseDown(int sx, int sy) {
                if (state==NORMAL && ic!=null) {
                        state = MOVING;
                        startX = ic.offScreenX(sx);
                        startY = ic.offScreenY(sy);
                        showStatus();
                }
        }

        protected void handleMouseUp(int screenX, int screenY) {
                state = NORMAL;
                if (imp==null) return;
                imp.draw(clipX-5, clipY-5, clipWidth+10, clipHeight+10);
                if (Recorder.record) {
                        String method;
                        if (type==LINE) {
                                Line line = (Line)imp.getRoi();
                                Recorder.record("makeLine", line.x1, line.y1, line.x2, line.y2);
                        } else if (type==OVAL)
                                Recorder.record("makeOval", x, y, width, height);
                        else if (!(this instanceof TextRoi)) {
                                if (cornerDiameter==0)
                                        Recorder.record("makeRectangle", x, y, width, height);
                                else {
                                        if (Recorder.scriptMode())
                                                Recorder.recordCall("imp.setRoi(new Roi("+x+", "+y+", "+width+", "+height+", "+cornerDiameter+"));");
                                        else
                                                Recorder.record("makeRectangle", x, y, width, height, cornerDiameter);
                                }
                        }
                }
                if (Toolbar.getToolId()==Toolbar.OVAL&&Toolbar.getBrushSize()>0)  {
                        int flags = ic!=null?ic.getModifiers():16;
                        if ((flags&16)==0) // erase ROI Brush
                                {imp.draw(); return;}
                }
                modifyRoi();
        }

    void modifyRoi() {
        if (previousRoi==null || previousRoi.modState==NO_MODS || imp==null)
                return;
                //IJ.log("modifyRoi: "+ type+"  "+modState+" "+previousRoi.type+"  "+previousRoi.modState);
        if (type==POINT || previousRoi.getType()==POINT) {
                if (type==POINT && previousRoi.getType()==POINT)
                        addPoint();
                else if (isArea() && previousRoi.getType()==POINT && previousRoi.modState==SUBTRACT_FROM_ROI)
                        subtractPoints();
                return;
        }
                Roi previous = (Roi)previousRoi.clone();
                previous.modState = NO_MODS;
        ShapeRoi s1  = null;
        ShapeRoi s2 = null;
        if (previousRoi instanceof ShapeRoi)
            s1 = (ShapeRoi)previousRoi;
        else
            s1 = new ShapeRoi(previousRoi);
        if (this instanceof ShapeRoi)
            s2 = (ShapeRoi)this;
        else
            s2 = new ShapeRoi(this);
        if (previousRoi.modState==ADD_TO_ROI)
                s1.or(s2);
        else
                s1.not(s2);
                previousRoi.modState = NO_MODS;
                Roi[] rois = s1.getRois();
                if (rois.length==0) return;
                int type2 = rois[0].getType();
                //IJ.log(rois.length+" "+type2);
                Roi roi2 = null;
                if (rois.length==1 && (type2==POLYGON||type2==FREEROI))
                        roi2 = rois[0];
                else
                        roi2 = s1;
                if (roi2!=null)
                        roi2.copyAttributes(previousRoi);
                imp.setRoi(roi2);
                previousRoi = previous;
    }

    void addPoint() {
                if (!(type==POINT && previousRoi.getType()==POINT)) {
                        modState = NO_MODS;
                        imp.draw();
                        return;
                }
                previousRoi.modState = NO_MODS;
                PointRoi p1 = (PointRoi)previousRoi;
                Rectangle r = getBounds();
                FloatPolygon poly = getFloatPolygon();
                imp.setRoi(p1.addPoint(poly.xpoints[0], poly.ypoints[0]));
    }

    void subtractPoints() {
                previousRoi.modState = NO_MODS;
                PointRoi p1 = (PointRoi)previousRoi;
                PointRoi p2 = p1.subtractPoints(this);
                if (p2!=null)
                        imp.setRoi(p1.subtractPoints(this));
                else
                        imp.killRoi();
    }

    /** If 'add' is true, adds this selection to the previous one. If 'subtract' is true, subtracts
        it from the previous selection. Called by the IJ.doWand() method, and the makeRectangle(),
        makeOval(), makePolygon() and makeSelection() macro functions. */
    public void update(boolean add, boolean subtract) {
        if (previousRoi==null) return;
        if (add) {
                        previousRoi.modState = ADD_TO_ROI;
                        modifyRoi();
                } else if (subtract) {
                        previousRoi.modState = SUBTRACT_FROM_ROI;
                        modifyRoi();
                } else
                        previousRoi.modState = NO_MODS;
     }

        protected void showStatus() {
                if (imp==null) return;
                String value;
                if (state!=CONSTRUCTING && (type==RECTANGLE||type==POINT) && width<=25 && height<=25) {
                        ImageProcessor ip = imp.getProcessor();
                        double v = ip.getPixelValue(x,y);
                        int digits = (imp.getType()==ImagePlus.GRAY8||imp.getType()==ImagePlus.GRAY16)?0:2;
                        value = ", value="+IJ.d2s(v,digits);
                } else
                        value = "";
                Calibration cal = imp.getCalibration();
                String size;
                if (cal.scaled() && !IJ.altKeyDown())
                        size = ", w="+IJ.d2s(width*cal.pixelWidth)+", h="+IJ.d2s(height*cal.pixelHeight);
                else
                        size = ", w="+width+", h="+height;
                size += ", ar="+IJ.d2s((double)width/height,2);
                IJ.showStatus(imp.getLocationAsString(x,y)+size+value);
        }

        /** Always returns null for rectangular Roi's */
        public ImageProcessor getMask() {
                if (cornerDiameter>0)
                        return (new ShapeRoi(new RoundRectangle2D.Float(x, y, width, height, cornerDiameter, cornerDiameter))).getMask();
                else
                        return null;
        }


        /**
        * @deprecated
        * replaced by setStrokeColor()
        */
        public void setInstanceColor(Color c) {
                 strokeColor = c;
        }

        /**
        * @deprecated
        * replaced by setStrokeWidth(int)
        */
        public void setLineWidth(int width) {
                setStrokeWidth(width) ;
        }

        public void updateWideLine(float width) {
                //IJ.log("updateWideLine "+isLine()+"  "+isDrawingTool()+"  "+getType());
                if (isLine()) {
                        wideLine = true;
                        setStrokeWidth(width);
                        if (getStrokeColor()==null) {
                                Color c = getColor();
                                setStrokeColor(new Color(c.getRed(),c.getGreen(),c.getBlue(), 77));
                        }
                }
        }

    /** Set 'nonScalable' true to have TextRois in a display
                list drawn at a fixed location and size. */
        public void setNonScalable(boolean nonScalable) {
                this.nonScalable = nonScalable;
        }


        protected BasicStroke getScaledStroke() {
                if (ic==null) return stroke;
                double mag = ic.getMagnification();
                if (mag!=1.0) {
                        float width = stroke.getLineWidth();
                        return new BasicStroke((float)(width*mag), BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL);
                } else
                        return stroke;
        }


        /** Sets the rounded rectangle corner diameter (pixels). */
        public void setCornerDiameter(int cornerDiameter) {
                if (cornerDiameter<0) cornerDiameter = 0;
                this.cornerDiameter = cornerDiameter;
                ImagePlus imp = WindowManager.getCurrentImage();
                if (imp!=null && this==imp.getRoi())
                        imp.updateAndDraw();
        }

        /** Returns the rounded rectangle corner diameter (pixels). */
        public int getCornerDiameter() {
                return cornerDiameter;
        }

        /** Obsolete; replaced by setCornerDiameter(). */
        public void setRoundRectArcSize(int cornerDiameter) {
                setCornerDiameter(cornerDiameter);
        }

        /** Obsolete; replaced by getCornerDiameter(). */
        public int getRoundRectArcSize() {
                return cornerDiameter;
        }


        // Used by the FileSaver and RoiEncoder to save overlay settings. */
        public void setPrototypeOverlay(Overlay overlay) {
                prototypeOverlay = new Overlay();
                prototypeOverlay.drawLabels(overlay.getDrawLabels());
                prototypeOverlay.drawNames(overlay.getDrawNames());
                prototypeOverlay.drawBackgrounds(overlay.getDrawBackgrounds());
                prototypeOverlay.setLabelColor(overlay.getLabelColor());
                prototypeOverlay.setLabelFont(overlay.getLabelFont());
        }

        // Used by the FileOpener and RoiDecoder to restore overlay settings. */
        public Overlay getPrototypeOverlay() {
                if (prototypeOverlay!=null)
                        return prototypeOverlay;
                else
                        return new Overlay();
        }

         public void startPaste(ImagePlus clipboard) {
                 IJ.showStatus("Pasting...");
                 this.clipboard = clipboard;
                 imp.getProcessor().snapshot();
                 updateClipRect();
                 imp.draw(clipX, clipY, clipWidth, clipHeight);
         }

         void updatePaste() {
                 if (clipboard!=null) {
                         imp.getMask();
                         ImageProcessor ip = imp.getProcessor();
                         ip.reset();
                         ip.copyBits(clipboard.getProcessor(), x, y, pasteMode);
                         if (type!=RECTANGLE)
                                 ip.reset(ip.getMask());
                         if (ic!=null)
                                 ic.setImageUpdated();
                 }
         }

         public void endPaste() {
                 if (clipboard!=null) {
                         updatePaste();
                         clipboard = null;
                         Undo.setup(Undo.FILTER, imp);
                 }
         }

         public void abortPaste() {
                 clipboard = null;
                 imp.getProcessor().reset();
                 imp.updateAndDraw();
         }

        /** Sets the Paste transfer mode.
                @see ij.process.Blitter
        */
        public static void setPasteMode(int transferMode) {
                if (transferMode==pasteMode) return;
                pasteMode = transferMode;
                ImagePlus imp = WindowManager.getCurrentImage();
                if (imp!=null)
                        imp.updateAndDraw();
        }

        /** Returns the current paste transfer mode, or NOT_PASTING (-1)
                if no paste operation is in progress.
                @see ij.process.Blitter
        */
        public int getPasteMode() {
                if (clipboard==null)
                        return NOT_PASTING;
                else
                        return pasteMode;
        }

        /** Returns the current paste transfer mode. */
        public static int getCurrentPasteMode() {
                return pasteMode;
        }

        /** Returns 'true' if this is an area selection. */
        public boolean isArea() {
                return (type>=RECTANGLE && type<=TRACED_ROI) || type==COMPOSITE;
        }

        /** Returns 'true' if this is a line selection. */
    public boolean isLine() {
        return type>=LINE && type<=FREELINE;
    }

        /** Returns 'true' if this is an ROI primarily used from drawing
                (e.g., TextRoi or Arrow). */
    public boolean isDrawingTool() {
        //return cornerDiameter>0;
        return false;
    }

    protected double getMagnification() {
        return ic!=null?ic.getMagnification():1.0;
    }

        /** Convenience method that converts Roi type to a human-readable form. */
        public String getTypeAsString() {
                String s="";
                switch(type) {
                        case POLYGON: s="Polygon"; break;
                        case FREEROI: s="Freehand"; break;
                        case TRACED_ROI: s="Traced"; break;
                        case POLYLINE: s="Polyline"; break;
                        case FREELINE: s="Freeline"; break;
                        case ANGLE: s="Angle"; break;
                        case LINE: s="Straight Line"; break;
                        case OVAL: s="Oval"; break;
                        case COMPOSITE: s = "Composite"; break;
                        case POINT: s = "Point"; break;
                        default: s="Rectangle"; break;
                }
                return s;
        }

        /** Returns true if this ROI is currently displayed on an image. */
        public boolean isVisible() {
                return ic!=null;
        }

        /** Returns true if this is a PolygonRoi that supports sub-pixel resolution. */
        public boolean subPixelResolution() {
                return subPixel;
        }

        /** Returns true if this is a PolygonRoi that supports sub-pixel
                resolution and polygons are drawn on zoomed images offset
                down and to the right by 0.5 pixels.. */
        public boolean getDrawOffset() {
                return false;
        }

        public void setDrawOffset(boolean drawOffset) {
        }

        /** Returns a copy of this roi. See Thinking is Java by Bruce Eckel
                (www.eckelobjects.com) for a good description of object cloning. */
        public synchronized Object clone() {
                try {
                        Roi r = (Roi)super.clone();
                        r.setImage(null);
                        r.setStroke(getStroke());
                        r.setFillColor(getFillColor());
                        r.imageID = getImageID();
                        return r;
                }
                //catch (CloneNotSupportedException e) {return null;}
                catch (Exception e) {return null;}
        }

        /** Checks whether two rectangles are equal. */
        public boolean equals(Object obj) {
                    if (obj instanceof Roi) {
                            Roi roi2 = (Roi)obj;
                            if (type!=roi2.getType()) return false;
                            if (!getBounds().equals(roi2.getBounds())) return false;
                            if (getLength()!=roi2.getLength()) return false;
                            return true;
                    } else
                            return false;
        }

        public String toString() {
                return ("Roi["+getTypeAsString()+", x="+x+", y="+y+", width="+width+", height="+height+"]");
        }
        protected int screenX(int ox) {return ic!=null?ic.screenX(ox):ox;}
        protected int screenY(int oy) {return ic!=null?ic.screenY(oy):oy;}
        protected int screenXD(double ox) {return ic!=null?ic.screenXD(ox):(int)ox;}
        protected int screenYD(double oy) {return ic!=null?ic.screenYD(oy):(int)oy;}

        /** Converts a float array to an int array using truncation. */
        public static int[] toInt(float[] arr) {
                return toInt(arr, null, arr.length);
        }

        public static int[] toInt(float[] arr, int[] arr2, int size) {
                int n = arr.length;
                if (size>n) size=n;
                int[] temp = arr2;
                if (temp==null || temp.length<n)
                        temp = new int[n];
                for (int i=0; i<size; i++)
                        temp[i] = (int)arr[i];
                        //temp[i] = (int)Math.floor(arr[i]+0.5);
                return temp;
        }

        /** Converts a float array to an int array using rounding. */
        public static int[] toIntR(float[] arr) {
                int n = arr.length;
                int[] temp = new int[n];
                for (int i=0; i<n; i++)
                        temp[i] = (int)Math.floor(arr[i]+0.5);
                return temp;
        }

        /** Converts an int array to a float array. */
        public static float[] toFloat(int[] arr) {
                int n = arr.length;
                float[] temp = new float[n];
                for (int i=0; i<n; i++)
                        temp[i] = arr[i];
                return temp;
        }


        /** @deprecated */
        public void temporarilyHide() {
        }

}
